#!/usr/bin/env python3

from pydoc import locate

from agora.abc import ProcessABC


class PostProcessABC(ProcessABC):
    """
    Extend ProcessABC to add as_function, allowing for all PostProcesses to be called as functions
    almost directly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def as_function(cls, data, *extra_data, **kwargs):
        # FIXME can this be a __call__ method instead?
        # Find the parameter's default
        parameters = cls.default_parameters(**kwargs)
        return cls(parameters=parameters).run(data, *extra_data)

    @classmethod
    def default_parameters(cls, *args, **kwargs):
        return get_parameters(cls.__name__).default(*args, **kwargs)


def get_process(process, suffix=""):
    """
    Dynamically import a process class from the available process locations.
    Assumes process filename and class name are the same

    Processes return the same shape as their input.
    MultiSignal either take or return multiple datasets (or both).
    Reshapers return a different shape for processes: Merger and Picker belong here.

    suffix : str
        Name of suffix, generally "" (empty) or "Parameters".
    """
    base_location = "postprocessor.core"
    possible_locations = ("processes", "multisignal", "reshapers")

    found = None
    for possible_location in possible_locations:
        location = (
            f"{base_location}.{possible_location}.{process}.{process}{suffix}"
        )
        found = locate(location)
        if found is not None:
            return found

    raise Exception(
        f"{process} not found in locations {possible_locations} at {base_location}"
    )


def get_parameters(process):
    """
    Dynamically import parameters from the 'processes' folder.
    Assumes parameter is the same name as the file with 'Parameters' added at the end.
    """
    return get_process(process, suffix="Parameters")



"""
Pipeline and chaining elements.
"""
import logging
import os
import re
import traceback
import typing as t
from copy import copy
from importlib.metadata import version
from pathlib import Path, PosixPath

import h5py
import numpy as np
import pandas as pd
from pathos.multiprocessing import Pool
from tqdm import tqdm

from agora.abc import ParametersABC, ProcessABC
from agora.io.metadata import MetaData, parse_logfiles
from agora.io.reader import StateReader
from agora.io.signal import Signal
from agora.io.writer import (
    LinearBabyWriter,
    StateWriter,
    TilerWriter,
)
from aliby.baby_client import BabyParameters, BabyRunner
from aliby.haystack import initialise_tf
from aliby.io.dataset import dispatch_dataset
from aliby.io.image import get_image_class
from aliby.tile.tiler import Tiler, TilerParameters
from extraction.core.extractor import Extractor, ExtractorParameters
from extraction.core.functions.defaults import exparams_from_meta
from postprocessor.core.processor import PostProcessor, PostProcessorParameters


class PipelineParameters(ParametersABC):
    """
    Parameters that host what is run and how. It takes a list of dictionaries, one for
    general in collection:
    pass dictionary for each step
    --------------------
    expt_id: int or str Experiment id (if integer) or local path (if string).
    directory: str Directory into which results are dumped. Default is "../data"

    Provides default parameters for the entire pipeline. This downloads the logfiles and sets the default
    timepoints and extraction parameters from there.
    """

    _pool_index = None

    def __init__(
        self, general, tiler, baby, extraction, postprocessing, reporting
    ):
        self.general = general
        self.tiler = tiler
        self.baby = baby
        self.extraction = extraction
        self.postprocessing = postprocessing
        self.reporting = reporting

    @classmethod
    def default(
        cls,
        general={},
        tiler={},
        baby={},
        extraction={},
        postprocessing={},
    ):
        expt_id = general.get("expt_id", 19993)
        if isinstance(expt_id, PosixPath):
            expt_id = str(expt_id)
            general["expt_id"] = expt_id

        directory = Path(general.get("directory", "../data"))

        with dispatch_dataset(
            expt_id,
            **{k: general.get(k) for k in ("host", "username", "password")},
        ) as conn:
            directory = directory / conn.unique_name
            if not directory.exists():
                directory.mkdir(parents=True)
                # Download logs to use for metadata
            conn.cache_logs(directory)
        try:
            meta_d = MetaData(directory, None).load_logs()
        except Exception as e:
            logging.getLogger("aliby").warn(
                f"WARNING:Metadata: error when loading: {e}"
            )
            minimal_default_meta = {
                "channels": ["Brightfield"],
                "ntps": [2000],
            }
            # Set minimal metadata
            meta_d = minimal_default_meta

        tps = meta_d.get("ntps", 2000)
        defaults = {
            "general": dict(
                id=expt_id,
                distributed=0,
                tps=tps,
                directory=str(directory.parent),
                filter="",
                earlystop=dict(
                    min_tp=100,
                    thresh_pos_clogged=0.4,
                    thresh_trap_ncells=8,
                    thresh_trap_area=0.9,
                    ntps_to_eval=5,
                ),
                logfile_level="INFO",
                use_explog=True,
            )
        }

        for k, v in general.items():  # Overwrite general parameters
            if k not in defaults["general"]:
                defaults["general"][k] = v
            elif isinstance(v, dict):
                for k2, v2 in v.items():
                    defaults["general"][k][k2] = v2
            else:
                defaults["general"][k] = v

        defaults["tiler"] = TilerParameters.default(**tiler).to_dict()
        defaults["baby"] = BabyParameters.default(**baby).to_dict()
        defaults["extraction"] = (
            exparams_from_meta(meta_d)
            or BabyParameters.default(**extraction).to_dict()
        )
        defaults["postprocessing"] = {}
        defaults["reporting"] = {}

        defaults["postprocessing"] = PostProcessorParameters.default(
            **postprocessing
        ).to_dict()
        defaults["reporting"] = {}

        return cls(**{k: v for k, v in defaults.items()})

    def load_logs(self):
        parsed_flattened = parse_logfiles(self.log_dir)
        return parsed_flattened


class Pipeline(ProcessABC):
    """
    A chained set of Pipeline elements connected through pipes.
    Tiling, Segmentation,Extraction and Postprocessing should use their own default parameters.
    These can be overriden passing the key:value of parameters to override to a PipelineParameters class

    """

    iterative_steps = ["tiler", "baby", "extraction"]

    step_sequence = [
        "tiler",
        "baby",
        "extraction",
        "postprocessing",
    ]

    # Indicate step-writer groupings to perform special operations during step iteration
    writer_groups = {
        "tiler": ["trap_info"],
        "baby": ["cell_info"],
        "extraction": ["extraction"],
        "postprocessing": ["postprocessing", "modifiers"],
    }
    writers = {  # TODO integrate Extractor and PostProcessing in here
        "tiler": [("tiler", TilerWriter)],
        "baby": [("baby", LinearBabyWriter), ("state", StateWriter)],
    }

    def __init__(self, parameters: PipelineParameters, store=None):
        super().__init__(parameters)

        if store is not None:
            store = Path(store)
        self.store = store

    @staticmethod
    def setLogger(
        folder, file_level: str = "INFO", stream_level: str = "WARNING"
    ):

        logger = logging.getLogger("aliby")
        logger.setLevel(getattr(logging, file_level))
        formatter = logging.Formatter(
            "%(asctime)s - %(levelname)s:%(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
        )

        ch = logging.StreamHandler()
        ch.setLevel(getattr(logging, stream_level))
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        # create file handler which logs even debug messages
        fh = logging.FileHandler(Path(folder) / "aliby.log", "w+")
        fh.setLevel(getattr(logging, file_level))
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    @classmethod
    def from_yaml(cls, fpath):
        # This is just a convenience function, think before implementing
        # for other processes
        return cls(parameters=PipelineParameters.from_yaml(fpath))

    @classmethod
    def from_folder(cls, dir_path):
        """
        Constructor to re-process all files in a given folder.

        Assumes all files share the same parameters (even if they don't share
        the same channel set).

        Parameters
        ---------
        dir_path : str or Pathlib indicating the folder containing the files to process
        """
        dir_path = Path(dir_path)
        files = list(dir_path.rglob("*.h5"))
        assert len(files), "No valid files found in folder"
        fpath = files[0]

        # TODO add support for non-standard unique folder names
        with h5py.File(fpath, "r") as f:
            pipeline_parameters = PipelineParameters.from_yaml(
                f.attrs["parameters"]
            )
        pipeline_parameters.general["directory"] = dir_path.parent
        pipeline_parameters.general["filter"] = [fpath.stem for fpath in files]

        # Fix legacy postprocessing parameters
        post_process_params = pipeline_parameters.postprocessing.get(
            "parameters", None
        )
        if post_process_params:
            pipeline_parameters.postprocessing["param_sets"] = copy(
                post_process_params
            )
            del pipeline_parameters.postprocessing["parameters"]

        return cls(pipeline_parameters)

    @classmethod
    def from_existing_h5(cls, fpath):
        """
        Constructor to process an existing hdf5 file.
        Notice that it forces a single file, not suitable for multiprocessing of certain positions.

        It i s also used as a base for a folder-wide reprocessing.
        """
        with h5py.File(fpath, "r") as f:
            pipeline_parameters = PipelineParameters.from_yaml(
                f.attrs["parameters"]
            )
        directory = Path(fpath).parent
        pipeline_parameters.general["directory"] = directory
        pipeline_parameters.general["filter"] = Path(fpath).stem

        post_process_params = pipeline_parameters.postprocessing.get(
            "parameters", None
        )
        if post_process_params:
            pipeline_parameters.postprocessing["param_sets"] = copy(
                post_process_params
            )
            del pipeline_parameters.postprocessing["parameters"]

        return cls(pipeline_parameters, store=directory)

    @property
    def _logger(self):
        return logging.getLogger("aliby")

    def run(self):
        """
        Config holds the general information, use in main
        Steps: all holds general tasks
        steps: strain_name holds task for a given strain
        """

        config = self.parameters.to_dict()
        expt_id = config["general"]["id"]
        distributed = config["general"]["distributed"]
        pos_filter = config["general"]["filter"]
        root_dir = Path(config["general"]["directory"])
        self.server_info = {
            k: config["general"].get(k)
            for k in ("host", "username", "password")
        }

        dispatcher = dispatch_dataset(expt_id, **self.server_info)
        logging.getLogger("aliby").info(
            f"Fetching data using {dispatcher.__class__.__name__}"
        )
        # Do all all initialisations

        with dispatcher as conn:
            image_ids = conn.get_images()

            directory = self.store or root_dir / conn.unique_name

            if not directory.exists():
                directory.mkdir(parents=True)

            # Download logs to use for metadata
            conn.cache_logs(directory)

        # Modify to the configuration
        self.parameters.general["directory"] = str(directory)
        config["general"]["directory"] = directory

        self.setLogger(directory)

        # Filter TODO integrate filter onto class and add regex
        def filt_int(d: dict, filt: int):
            return {k: v for i, (k, v) in enumerate(d.items()) if i == filt}

        def filt_str(image_ids: dict, filt: str):
            return {k: v for k, v in image_ids.items() if re.search(filt, k)}

        def pick_filter(image_ids: dict, filt: int or str):
            if isinstance(filt, str):
                image_ids = filt_str(image_ids, filt)
            elif isinstance(filt, int):
                image_ids = filt_int(image_ids, filt)
            return image_ids

        if isinstance(pos_filter, list):
            image_ids = {
                k: v
                for filt in pos_filter
                for k, v in pick_filter(image_ids, filt).items()
            }
        else:
            image_ids = pick_filter(image_ids, pos_filter)

        assert len(image_ids), "No images to segment"

        if distributed != 0:  # Gives the number of simultaneous processes
            with Pool(distributed) as p:
                results = p.map(
                    lambda x: self.create_pipeline(*x),
                    [(k, i) for i, k in enumerate(image_ids.items())],
                    # num_cpus=distributed,
                    # position=0,
                )

        else:  # Sequential
            results = []
            for k, v in tqdm(image_ids.items()):
                r = self.create_pipeline((k, v), 1)
                results.append(r)

        return results

    def create_pipeline(
        self,
        image_id: t.Tuple[str, str or PosixPath or int],
        index: t.Optional[int] = None,
    ):
        """ """
        self._pool_index = index
        name, image_id = image_id
        session = None
        filename = None
        run_kwargs = {"extraction": {"labels": None, "masks": None}}
        try:
            (
                filename,
                meta,
                config,
                process_from,
                tps,
                steps,
                earlystop,
                session,
                trackers_state,
            ) = self._setup_pipeline(image_id)

            loaded_writers = {
                name: writer(filename)
                for k in self.step_sequence
                if k in self.writers
                for name, writer in self.writers[k]
            }
            writer_ow_kwargs = {
                "state": loaded_writers["state"].datatypes.keys(),
                "baby": ["mother_assign"],
            }

            # START PIPELINE
            frac_clogged_traps = 0
            min_process_from = min(process_from.values())

            with get_image_class(image_id)(
                image_id, **self.server_info
            ) as image:

                # Initialise Steps
                if "tiler" not in steps:
                    steps["tiler"] = Tiler.from_image(
                        image, TilerParameters.from_dict(config["tiler"])
                    )

                if process_from["baby"] < tps:
                    session = initialise_tf(2)
                    steps["baby"] = BabyRunner.from_tiler(
                        BabyParameters.from_dict(config["baby"]),
                        steps["tiler"],
                    )
                    if trackers_state:
                        steps["baby"].crawler.tracker_states = trackers_state

                # Limit extraction parameters during run using the available channels in tiler
                if process_from["extraction"] < tps:
                    # TODO Move this parameter validation into Extractor
                    av_channels = set((*steps["tiler"].channels, "general"))
                    config["extraction"]["tree"] = {
                        k: v
                        for k, v in config["extraction"]["tree"].items()
                        if k in av_channels
                    }
                    config["extraction"]["sub_bg"] = av_channels.intersection(
                        config["extraction"]["sub_bg"]
                    )

                    av_channels_wsub = av_channels.union(
                        [c + "_bgsub" for c in config["extraction"]["sub_bg"]]
                    )
                    tmp = copy(config["extraction"]["multichannel_ops"])
                    for op, (input_ch, _, _) in tmp.items():
                        if not set(input_ch).issubset(av_channels_wsub):
                            del config["extraction"]["multichannel_ops"][op]

                    exparams = ExtractorParameters.from_dict(
                        config["extraction"]
                    )
                    steps["extraction"] = Extractor.from_tiler(
                        exparams, store=filename, tiler=steps["tiler"]
                    )
                    pbar = tqdm(
                        range(min_process_from, tps),
                        desc=image.name,
                        initial=min_process_from,
                        total=tps,
                        # position=index + 1,
                    )
                    for i in pbar:

                        if (
                            frac_clogged_traps
                            < earlystop["thresh_pos_clogged"]
                            or i < earlystop["min_tp"]
                        ):

                            for step in self.iterative_steps:
                                if i >= process_from[step]:
                                    result = steps[step].run_tp(
                                        i, **run_kwargs.get(step, {})
                                    )
                                    if step in loaded_writers:
                                        loaded_writers[step].write(
                                            data=result,
                                            overwrite=writer_ow_kwargs.get(
                                                step, []
                                            ),
                                            tp=i,
                                            meta={"last_processed": i},
                                        )

                                    # Step-specific actions
                                    if (
                                        step == "tiler"
                                        and i == min_process_from
                                    ):
                                        logging.getLogger("aliby").info(
                                            f"Found {steps['tiler'].n_traps} traps in {image.name}"
                                        )
                                    elif (
                                        step == "baby"
                                    ):  # Write state and pass info to ext
                                        loaded_writers["state"].write(
                                            data=steps[
                                                step
                                            ].crawler.tracker_states,
                                            overwrite=loaded_writers[
                                                "state"
                                            ].datatypes.keys(),
                                            tp=i,
                                        )
                                    elif (
                                        step == "extraction"
                                    ):  # Remove mask/label after ext
                                        for k in ["masks", "labels"]:
                                            run_kwargs[step][k] = None

                            frac_clogged_traps = self.check_earlystop(
                                filename, earlystop, steps["tiler"].tile_size
                            )
                            self._log(
                                f"{name}:Clogged_traps:{frac_clogged_traps}"
                            )

                            frac = np.round(frac_clogged_traps * 100)
                            pbar.set_postfix_str(f"{frac} Clogged")
                        else:  # Stop if more than X% traps are clogged
                            self._log(
                                f"{name}:Analysis stopped early at time {i} with {frac_clogged_traps} clogged traps"
                            )
                            meta.add_fields({"end_status": "Clogged"})
                            break

                        meta.add_fields({"last_processed": i})

                    # Run post-processing
                    meta.add_fields({"end_status": "Success"})
                    post_proc_params = PostProcessorParameters.from_dict(
                        config["postprocessing"]
                    )
                    PostProcessor(filename, post_proc_params).run()

                    self._log("Analysis finished successfully.", "info")
                    return 1

        except Exception as e:  # Catch bugs during setup or runtime
            logging.exception(
                f"{name}: Exception caught.",
                exc_info=True,
            )
            # This prints the type, value, and stack trace of the
            # current exception being handled.
            traceback.print_exc()
            raise e
        finally:
            _close_session(session)

    @staticmethod
    def check_earlystop(filename: str, es_parameters: dict, tile_size: int):
        s = Signal(filename)
        df = s["/extraction/general/None/area"]
        cells_used = df[
            df.columns[-1 - es_parameters["ntps_to_eval"] : -1]
        ].dropna(how="all")
        traps_above_nthresh = (
            cells_used.groupby("trap").count().apply(np.mean, axis=1)
            > es_parameters["thresh_trap_ncells"]
        )
        traps_above_athresh = (
            cells_used.groupby("trap").sum().apply(np.mean, axis=1)
            / tile_size**2
            > es_parameters["thresh_trap_area"]
        )

        return (traps_above_nthresh & traps_above_athresh).mean()

    def _load_config_from_file(
        self,
        filename: PosixPath,
        process_from: t.Dict[str, int],
        trackers_state: t.List,
        overwrite: t.Dict[str, bool],
    ):
        with h5py.File(filename, "r") as f:
            for k in process_from.keys():
                if not overwrite[k]:
                    process_from[k] = self.legacy_get_last_tp[k](f)
                    process_from[k] += 1
        return process_from, trackers_state, overwrite

    @staticmethod
    def legacy_get_last_tp(step: str) -> t.Callable:
        """Get last time-point in different ways depending
        on which step we are using

        To support segmentation in aliby < v0.24
        TODO Deprecate and replace with State method
        """
        switch_case = {
            "tiler": lambda f: f["trap_info/drifts"].shape[0] - 1,
            "baby": lambda f: f["cell_info/timepoint"][-1],
            "extraction": lambda f: f[
                "extraction/general/None/area/timepoint"
            ][-1],
        }
        return switch_case[step]

    def _setup_pipeline(
        self, image_id: int
    ) -> t.Tuple[
        PosixPath,
        MetaData,
        t.Dict,
        int,
        t.Dict,
        t.Dict,
        t.Optional[int],
        t.List[np.ndarray],
    ]:
        """
        Initialise pipeline components and if necessary use
        exising file to continue existing experiments.


        Parameters
        ----------
        image_id : int
            identifier of image in OMERO server, or filename

        Returns
        ---------
        filename: str
        meta:
        config:
        process_from:
        tps:
        steps:
        earlystop:
        session:
        trackers_state:

        Examples
        --------
        FIXME: Add docs.

        """
        config = self.parameters.to_dict()
        pparams = config
        image_id = image_id
        general_config = config["general"]
        session = None
        earlystop = general_config.get("earlystop", None)
        process_from = {k: 0 for k in self.iterative_steps}
        steps = {}
        ow = {k: 0 for k in self.step_sequence}

        # check overwriting
        ow_id = general_config.get("overwrite", 0)
        ow = {step: True for step in self.step_sequence}
        if ow_id and ow_id is not True:
            ow = {
                step: self.step_sequence.index(ow_id) < i
                for i, step in enumerate(self.step_sequence, 1)
            }

        # Set up
        directory = general_config["directory"]

        trackers_state: t.List[np.ndarray] = []
        with get_image_class(image_id)(image_id, **self.server_info) as image:
            filename = Path(f"{directory}/{image.name}.h5")
            meta = MetaData(directory, filename)

            from_start = True if np.any(ow.values()) else False

            # New experiment or overwriting
            if (
                from_start
                and (
                    config.get("overwrite", False) == True
                    or np.all(list(ow.values()))
                )
                and filename.exists()
            ):
                os.remove(filename)

            # If no previous segmentation and keep tiler
            if filename.exists():
                self._log("Result file exists.", "info")
                if not ow["tiler"]:
                    steps["tiler"] = Tiler.from_hdf5(image, filename)
                    try:
                        (
                            process_from,
                            trackers_state,
                            ow,
                        ) = self._load_config_from_file(
                            filename, process_from, trackers_state, ow
                        )
                        # get state array
                        trackers_state = (
                            []
                            if ow["baby"]
                            else StateReader(filename).get_formatted_states()
                        )

                        config["tiler"] = steps["tiler"].parameters.to_dict()
                    except Exception:
                        pass

            if config["general"]["use_explog"]:
                meta.run()

            meta.add_fields(  # Add non-logfile metadata
                {
                    "aliby_version": version("aliby"),
                    "baby_version": version("aliby-baby"),
                    "omero_id": config["general"]["id"],
                    "image_id": image_id
                    if isinstance(image_id, int)
                    else str(image_id),
                    "parameters": PipelineParameters.from_dict(
                        pparams
                    ).to_yaml(),
                }
            )

            tps = min(general_config["tps"], image.data.shape[0])

            return (
                filename,
                meta,
                config,
                process_from,
                tps,
                steps,
                earlystop,
                session,
                trackers_state,
            )


def _close_session(session):
    if session:
        session.close()



#!/usr/bin/env jupyter


def run():
    import argparse

    from aliby.pipeline import Pipeline, PipelineParameters

    parser = argparse.ArgumentParser(
        prog="aliby-run",
        description="Run a default microscopy analysis pipeline",
    )

    param_values = {
        "expt_id": None,
        "distributed": 2,
        "tps": 2,
        "directory": "./data",
        "filter": 0,
        "host": None,
        "username": None,
        "password": None,
    }

    def _cast_str(x: str or None):
        """
        Cast string as int if possible. If Nonetype return None.
        """
        if x:
            try:
                return int(x)
            except:
                return x

    for k in param_values:
        parser.add_argument(f"--{k}", action="store")

    args = parser.parse_args()

    for k in param_values:
        if passed_value := _cast_str(getattr(args, k)):

            param_values[k] = passed_value

    params = PipelineParameters.default(general=param_values)
    p = Pipeline(params)

    p.run()



# TODO remove/to snippets?
"""
Post-processing utilities

Notes: I don't have statistics on ranges of radii for each of the knots in
the radial spline representation, but we regularly extract the average of
these radii for each cell. So, depending on camera/lens, we get:
    * 60x evolve: mean radii of 2-14 pixels (and measured areas of 30-750
    pixels^2)
    * 60x prime95b: mean radii of 3-24 pixels (and measured areas of 60-2000
    pixels^2)

And I presume that for a 100x lens we would get an ~5/3 increase over those
values.

In terms of the current volume estimation method, it's currently only
implemented in the AnalysisToolbox repository, but it's super simple:

mVol = 4/3*pi*sqrt(mArea/pi).^3

where mArea is simply the sum of pixels for that cell.


These functions were developed by D. Adjavon in 2020 to find the most accurate
way to calculate cell volume from a mask.

"""
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy import ndimage
from skimage import draw, measure
from skimage.morphology import ball, erosion


def my_ball(radius):
    """Generates a ball-shaped structuring element.

    This is the 3D equivalent of a disk.
    A pixel is within the neighborhood if the Euclidean distance between
    it and the origin is no greater than radius.

    Parameters
    ----------
    radius : int
        The radius of the ball-shaped structuring element.

    Other Parameters
    ----------------
    dtype : data-type
        The data type of the structuring element.

    Returns
    -------
    selem : ndarray
        The structuring element where elements of the neighborhood
        are 1 and 0 otherwise.
    """
    n = 2 * radius + 1
    Z, Y, X = np.mgrid[
        -radius : radius : n * 1j,
        -radius : radius : n * 1j,
        -radius : radius : n * 1j,
    ]
    X **= 2
    Y **= 2
    Z **= 2
    X += Y
    X += Z
    # s = X ** 2 + Y ** 2 + Z ** 2
    return X <= radius * radius


def circle_outline(r):
    return ellipse_perimeter(r, r)


def ellipse_perimeter(x, y):
    im_shape = int(2 * max(x, y) + 1)
    img = np.zeros((im_shape, im_shape), dtype=np.uint8)
    rr, cc = draw.ellipse_perimeter(
        int(im_shape // 2), int(im_shape // 2), int(x), int(y)
    )
    img[rr, cc] = 1
    return np.pad(img, 1)


def capped_cylinder(x, y):
    max_size = y + 2 * x + 2
    pixels = np.zeros((max_size, max_size))

    rect_start = ((max_size - x) // 2, x + 1)
    rr, cc = draw.rectangle_perimeter(
        rect_start, extent=(x, y), shape=(max_size, max_size)
    )
    pixels[rr, cc] = 1
    circle_centres = [
        (max_size // 2 - 1, x),
        (max_size // 2 - 1, max_size - x - 1),
    ]
    for r, c in circle_centres:
        rr, cc = draw.circle_perimeter(
            r, c, (x + 1) // 2, shape=(max_size, max_size)
        )
        pixels[rr, cc] = 1
    pixels = ndimage.morphology.binary_fill_holes(pixels)
    pixels ^= erosion(pixels)
    return pixels


def volume_of_sphere(radius):
    return 4 / 3 * np.pi * radius**3


def plot_voxels(voxels):
    verts, faces, _, _ = measure.marching_cubes_lewiner(voxels, 0)
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection="3d")
    mesh = Poly3DCollection(verts[faces])
    mesh.set_edgecolor("k")
    ax.add_collection3d(mesh)
    ax.set_xlim(0, voxels.shape[0])
    ax.set_ylim(0, voxels.shape[1])
    ax.set_zlim(0, voxels.shape[2])
    plt.tight_layout()
    plt.show()


# Volume estimation
def union_of_spheres(outline, shape="my_ball", debug=False):
    filled = ndimage.binary_fill_holes(outline)
    nearest_neighbor = (
        ndimage.morphology.distance_transform_edt(outline == 0) * filled
    )
    voxels = np.zeros((filled.shape[0], filled.shape[1], max(filled.shape)))
    c_z = voxels.shape[2] // 2
    for x, y in zip(*np.where(filled)):
        radius = nearest_neighbor[(x, y)]
        if radius > 0:
            if shape == "ball":
                b = ball(radius)
            elif shape == "my_ball":
                b = my_ball(radius)
            else:
                raise ValueError(
                    f"{shape} is not an accepted value for " f"shape."
                )
            centre_b = ndimage.measurements.center_of_mass(b)

            I, J, K = np.ogrid[: b.shape[0], : b.shape[1], : b.shape[2]]
            voxels[
                I + int(x - centre_b[0]),
                J + int(y - centre_b[1]),
                K + int(c_z - centre_b[2]),
            ] += b
    if debug:
        plot_voxels(voxels)
    return voxels.astype(bool).sum()


def improved_uos(outline, shape="my_ball", debug=False):
    filled = ndimage.binary_fill_holes(outline)
    nearest_neighbor = (
        ndimage.morphology.distance_transform_edt(outline == 0) * filled
    )
    voxels = np.zeros((filled.shape[0], filled.shape[1], max(filled.shape)))
    c_z = voxels.shape[2] // 2

    while np.any(nearest_neighbor != 0):
        radius = np.max(nearest_neighbor)
        x, y = np.argwhere(nearest_neighbor == radius)[0]
        if shape == "ball":
            b = ball(np.ceil(radius))
        elif shape == "my_ball":
            b = my_ball(np.ceil(radius))
        else:
            raise ValueError(f"{shape} is not an accepted value for shape")
        centre_b = ndimage.measurements.center_of_mass(b)

        I, J, K = np.ogrid[: b.shape[0], : b.shape[1], : b.shape[2]]
        voxels[
            I + int(x - centre_b[0]),
            J + int(y - centre_b[1]),
            K + int(c_z - centre_b[2]),
        ] += b

        # Use the central disk of the ball from voxels to get the circle
        # = 0 if nn[x,y] < r else nn[x,y]
        rr, cc = draw.circle(x, y, np.ceil(radius), nearest_neighbor.shape)
        nearest_neighbor[rr, cc] = 0
    if debug:
        plot_voxels(voxels)
    return voxels.astype(bool).sum()


def conical(outline, debug=False):
    nearest_neighbor = ndimage.morphology.distance_transform_edt(
        outline == 0
    ) * ndimage.binary_fill_holes(outline)
    if debug:
        hf = plt.figure()
        ha = hf.add_subplot(111, projection="3d")

        X, Y = np.meshgrid(
            np.arange(nearest_neighbor.shape[0]),
            np.arange(nearest_neighbor.shape[1]),
        )
        ha.plot_surface(X, Y, nearest_neighbor)
        plt.show()
    return 4 * nearest_neighbor.sum()


def volume(outline, method="spheres"):
    if method == "conical":
        return conical(outline)
    elif method == "spheres":
        return union_of_spheres(outline)
    else:
        raise ValueError(f"Method {method} not implemented.")



"""
Tiler: Tiles and tracks traps.

The tasks of the Tiler are selecting regions of interest, or tiles, of an image - with one tile per trap, tracking and correcting for the drift of the microscope stage over time, and handling errors and bridging between the image data and ALIBY’s image-processing steps.

Tiler subclasses deal with either network connections or local files.

To find traps, we use a two-step process: we analyse the bright-field image to produce the template of a trap, and we fit this template to the image to find the traps' centres.

We use texture-based segmentation (entropy) to split the image into foreground -- cells and traps -- and background, which we then identify with an Otsu filter. Two methods are used to produce a template trap from these regions: pick the trap with the smallest minor axis length and average over all validated traps.

A peak-identifying algorithm recovers the x and y-axis location of traps in the original image, and we choose the templating approach that identifies the most traps

One key method is Tiler.run.

The image-processing is performed by traps/segment_traps.

The experiment is stored as an array with a standard indexing order of (Time, Channels, Z-stack, X, Y).
"""
import re
import typing as t
import warnings
from functools import lru_cache
from pathlib import PosixPath

import dask.array as da
import h5py
import numpy as np
from skimage.registration import phase_cross_correlation

from agora.abc import ParametersABC, StepABC
from agora.io.writer import BridgeH5
from aliby.io.image import ImageLocalOME, ImageDir, ImageDummy
from aliby.tile.traps import segment_traps


class Trap:
    """
    Stores a trap's location and size.
    Allows checks to see if the trap should be padded.
    Can export the trap either in OMERO or numpy formats.
    """

    def __init__(self, centre, parent, size, max_size):
        self.centre = centre
        self.parent = parent  # used to access drifts
        self.size = size
        self.half_size = size // 2
        self.max_size = max_size

    def at_time(self, tp: int) -> t.List[int]:
        """
        Return trap centre at time tp by applying drifts

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        trap_centre:
        """
        drifts = self.parent.drifts
        trap_centre = self.centre - np.sum(drifts[: tp + 1], axis=0)
        return list(trap_centre.astype(int))

    def as_tile(self, tp):
        """
        Return trap in the OMERO tile format of x, y, w, h
        where x, y are at the bottom left corner of the tile
        and w and h are the tile width and height.

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        x: int
            x-coordinate of bottom left corner of tile
        y: int
            y-coordinate of bottom left corner of tile
        w: int
            Width of tile
        h: int
            Height of tile
        """
        x, y = self.at_time(tp)
        # tile bottom corner
        x = int(x - self.half_size)
        y = int(y - self.half_size)
        return x, y, self.size, self.size

    def as_range(self, tp):
        """
        Return trap in a range format: two slice objects that can
        be used in arrays

        Parameters
        ----------
        tp: integer
            Index for a time point

        Returns
        -------
        A slice of x coordinates from left to right
        A slice of y coordinates from top to bottom
        """
        x, y, w, h = self.as_tile(tp)
        return slice(x, x + w), slice(y, y + h)


class TrapLocations:
    """
    Stores each trap as an instance of Trap.
    Traps can be iterated.
    """

    def __init__(
        self,
        initial_location: np.array,
        tile_size: int = None,
        max_size: int = 1200,
        drifts: np.array = None,
    ):
        if drifts is None:
            drifts = []
        self.tile_size = tile_size
        self.max_size = max_size
        self.initial_location = initial_location
        self.traps = [
            Trap(centre, self, tile_size or max_size, max_size)
            for centre in initial_location
        ]
        self.drifts = drifts

    def __len__(self):
        return len(self.traps)

    def __iter__(self):
        yield from self.traps

    @property
    def shape(self):
        """
        Returns no of traps and no of drifts
        """
        return len(self.traps), len(self.drifts)

    def to_dict(self, tp):
        """
        Export inital locations, tile_size, max_size, and drifts
        as a dictionary

        Parameters
        ----------
        tp: integer
            An index for a time point
        """
        res = dict()
        if tp == 0:
            res["trap_locations"] = self.initial_location
            res["attrs/tile_size"] = self.tile_size
            res["attrs/max_size"] = self.max_size
        res["drifts"] = np.expand_dims(self.drifts[tp], axis=0)
        return res

    def at_time(self, tp: int) -> np.ndarray:
        # Returns ( ntraps, 2 ) ndarray with the trap centres as individual rows
        return np.array([trap.at_time(tp) for trap in self.traps])

    @classmethod
    def from_tiler_init(
        cls, initial_location, tile_size: int = None, max_size: int = 1200
    ):
        """
        Instantiate class from an instance of the Tiler class
        """
        return cls(initial_location, tile_size, max_size, drifts=[])

    @classmethod
    def read_hdf5(cls, file):
        """
        Instantiate class from a hdf5 file
        """
        with h5py.File(file, "r") as hfile:
            trap_info = hfile["trap_info"]
            initial_locations = trap_info["trap_locations"][()]
            drifts = trap_info["drifts"][()].tolist()
            max_size = trap_info.attrs["max_size"]
            tile_size = trap_info.attrs["tile_size"]
        trap_locs = cls(initial_locations, tile_size, max_size=max_size)
        trap_locs.drifts = drifts
        return trap_locs


class TilerParameters(ParametersABC):
    _defaults = {"tile_size": 117, "ref_channel": "Brightfield", "ref_z": 0}


class Tiler(StepABC):
    """
    Remote Timelapse Tiler.

    Finds traps and re-registers images if there is any drifting.
    Fetches images from a server.

    Uses an Image instance, which lazily provides the data on pixels, and, as
    an independent argument, metadata.
    """

    def __init__(
        self,
        image: da.core.Array,
        metadata: dict,
        parameters: TilerParameters,
        trap_locs=None,
    ):
        """
        Initialise Tiler

        Parameters
        ----------
        image: an instance of Image
        metadata: dictionary
        parameters: an instance of TilerPameters
        trap_locs: (optional)
        """
        super().__init__(parameters)
        self.image = image
        self._metadata = metadata
        self.channels = metadata.get(
            "channels", list(range(metadata["size_c"]))
        )
        self.ref_channel = self.get_channel_index(parameters.ref_channel)

        self.trap_locs = trap_locs
        try:
            self.z_perchannel = {
                ch: zsect
                for ch, zsect in zip(self.channels, metadata["zsections"])
            }
        except Exception as e:
            self._log(f"No z_perchannel data: {e}")

        self.tile_size = self.tile_size or min(self.image.shape[-2:])

    @classmethod
    def dummy(cls, parameters: dict):
        """
        Instantiate dummy Tiler from dummy image

        If image.dimorder exists dimensions are saved in that order.
        Otherwise default to "tczyx".

        Parameters
        ----------
        parameters: dictionary output of an instance of TilerParameters
        """
        imgdmy_obj = ImageDummy(parameters)
        dummy_image = imgdmy_obj.get_data_lazy()
        # Default to "tczyx" if image.dimorder is None
        dummy_omero_metadata = {
            f"size_{dim}": dim_size
            for dim, dim_size in zip(
                imgdmy_obj.dimorder or "tczyx", dummy_image.shape
            )
        }
        dummy_omero_metadata.update(
            {
                "channels": [
                    parameters["ref_channel"],
                    *(["nil"] * (dummy_omero_metadata["size_c"] - 1)),
                ],
                "name": "",
            }
        )

        return cls(
            imgdmy_obj.data,
            dummy_omero_metadata,
            TilerParameters.from_dict(parameters),
        )

    @classmethod
    def from_image(cls, image, parameters: TilerParameters):
        """
        Instantiate Tiler from an Image instance

        Parameters
        ----------
        image: an instance of Image
        parameters: an instance of TilerPameters
        """
        return cls(image.data, image.metadata, parameters)

    @classmethod
    def from_h5(
        cls,
        image: t.Union[
            ImageLocalOME, ImageDir
        ],  # TODO provide baseclass instead
        filepath: t.Union[str, PosixPath],
        parameters: TilerParameters = None,
    ):
        """
        Instantiate Tiler from hdf5 files

        Parameters
        ----------
        image: an instance of Image
        filepath: Path instance
            Path to a directory of h5 files
        parameters: an instance of TileParameters (optional)
        """
        trap_locs = TrapLocations.read_hdf5(filepath)
        metadata = BridgeH5(filepath).meta_h5
        metadata["channels"] = image.metadata["channels"]
        if parameters is None:
            parameters = TilerParameters.default()
        tiler = cls(
            image.data,
            metadata,
            parameters,
            trap_locs=trap_locs,
        )
        if hasattr(trap_locs, "drifts"):
            tiler.n_processed = len(trap_locs.drifts)
        return tiler

    @lru_cache(maxsize=2)
    def get_tc(self, t, c):
        """
        Load image using dask.
        Assumes the image is arranged as
            no of time points
            no of channels
            no of z stacks
            no of pixels in y direction
            no of pixels in x direction

        Parameters
        ----------
        t: integer
            An index for a time point
        c: integer
            An index for a channel

        Retruns
        -------
        full: an array of images
        """
        full = self.image[t, c].compute(scheduler="synchronous")
        return full

    @property
    def shape(self):
        """
        Returns properties of the time-lapse as shown by self.image.shape

        """
        return self.image.shape

    @property
    def n_processed(self):
        """
        Returns the number of images that have been processed
        """
        if not hasattr(self, "_n_processed"):
            self._n_processed = 0
        return self._n_processed

    @n_processed.setter
    def n_processed(self, value):
        self._n_processed = value

    @property
    def n_traps(self):
        """
        Returns number of traps
        """
        return len(self.trap_locs)

    def initialise_traps(self, tile_size: int = None):
        """
        Find initial trap positions if they have not been initialised.
        Removes all those that are too close to the edge so no padding
        is necessary.

        Parameters
        ----------
        tile_size: integer
            The size of a tile
        """
        initial_image = self.image[0, self.ref_channel, self.ref_z]
        if tile_size:
            half_tile = tile_size // 2
            # max_size is the minimal number of x or y pixels
            max_size = min(self.image.shape[-2:])
            # first time point, reference channel, reference z-position
            # find the traps
            trap_locs = segment_traps(initial_image, tile_size)
            # keep only traps that are not near an edge
            trap_locs = [
                [x, y]
                for x, y in trap_locs
                if half_tile < x < max_size - half_tile
                and half_tile < y < max_size - half_tile
            ]
            # store traps in an instance of TrapLocations
            self.trap_locs = TrapLocations.from_tiler_init(
                trap_locs, tile_size
            )
        else:
            yx_shape = self.image.shape[-2:]
            trap_locs = [[x // 2 for x in yx_shape]]
            self.trap_locs = TrapLocations.from_tiler_init(
                trap_locs, max_size=min(yx_shape)
            )

    def find_drift(self, tp):
        """
        Find any translational drift between two images at consecutive
        time points using cross correlation.

        Arguments
        ---------
        tp: integer
            Index for a time point
        """
        prev_tp = max(0, tp - 1)
        # cross-correlate
        drift, _, _ = phase_cross_correlation(
            self.image[prev_tp, self.ref_channel, self.ref_z],
            self.image[tp, self.ref_channel, self.ref_z],
        )
        # store drift
        if 0 < tp < len(self.trap_locs.drifts):
            self.trap_locs.drifts[tp] = drift.tolist()
        else:
            self.trap_locs.drifts.append(drift.tolist())

    def get_tp_data(self, tp, c):
        """
        Returns all traps corrected for drift.

        Parameters
        ----------
        tp: integer
            An index for a time point
        c: integer
            An index for a channel
        """
        traps = []
        # get OMERO image
        full = self.get_tc(tp, c)
        for trap in self.trap_locs:
            # pad trap if necessary
            ndtrap = self.ifoob_pad(full, trap.as_range(tp))
            traps.append(ndtrap)
        return np.stack(traps)

    def get_trap_data(self, trap_id, tp, c):
        """
        Returns a particular trap corrected for drift and padding

        Parameters
        ----------
        trap_id: integer
            Number of trap
        tp: integer
            Index of time points
        c: integer
            Index of channel

        Returns
        -------
        ndtrap: array
            An array of (x, y) arrays, one for each z stack
        """
        full = self.get_tc(tp, c)
        trap = self.trap_locs.traps[trap_id]
        ndtrap = self.ifoob_pad(full, trap.as_range(tp))
        return ndtrap

    def _run_tp(self, tp):
        """
        Find traps if they have not yet been found.
        Determine any translational drift of the current image from the
        previous one.

        Arguments
        ---------
        tp: integer
            The time point to tile.
        """
        # assert tp >= self.n_processed, "Time point already processed"
        # TODO check contiguity?
        if self.n_processed == 0 or not hasattr(self.trap_locs, "drifts"):
            self.initialise_traps(self.tile_size)
        if hasattr(self.trap_locs, "drifts"):
            drift_len = len(self.trap_locs.drifts)
            if self.n_processed != drift_len:
                warnings.warn("Tiler:n_processed and ndrifts don't match")
                self.n_processed = drift_len
        # determine drift
        self.find_drift(tp)
        # update n_processed
        self.n_processed = tp + 1
        # return result for writer
        return self.trap_locs.to_dict(tp)

    def run(self, time_dim=None):
        """
        Tile all time points in an experiment at once.
        """
        if time_dim is None:
            time_dim = 0
        for frame in range(self.image.shape[time_dim]):
            self.run_tp(frame)
        return None

    def get_traps_timepoint(self, *args, **kwargs):
        self._log(
            "get_trap_timepoints is deprecated; get_tiles_timepoint instead."
        )

        return self.get_tiles_timepoint(*args, **kwargs)

    # The next set of functions are necessary for the extraction object
    def get_tiles_timepoint(
        self, tp, tile_shape=None, channels=None, z: int = 0
    ) -> np.ndarray:
        """
        Get a multidimensional array with all tiles for a set of channels
        and z-stacks.

        Used by extractor.

        Parameters
        ---------
        tp: int
            Index of time point
        tile_shape: int or tuple of two ints
            Size of tile in x and y dimensions
        channels: string or list of strings
            Names of channels of interest
        z: int
            Index of z-channel of interest

        Returns
        -------
        res: array
            Data arranged as (traps, channels, timepoints, X, Y, Z)
        """
        # FIXME add support for subtiling trap
        # FIXME can we ignore z(always  give)
        if channels is None:
            channels = [0]
        elif isinstance(channels, str):
            channels = [channels]
        # get the data
        res = []
        for c in channels:
            # only return requested z
            val = self.get_tp_data(tp, c)[:, z]
            # starts with the order: traps, z, y, x
            # returns the order: trap, C, T, X, Y, Z
            val = val.swapaxes(1, 3).swapaxes(1, 2)
            val = np.expand_dims(val, axis=1)
            res.append(val)
        if tile_shape is not None:
            if isinstance(tile_shape, int):
                tile_shape = (tile_shape, tile_shape)
            assert np.all(
                [
                    (tile_size - ax) > -1
                    for tile_size, ax in zip(tile_shape, res[0].shape[-3:-2])
                ]
            )
        return np.stack(res, axis=1)

    @property
    def ref_channel_index(self):
        return self.get_channel_index(self.parameters.ref_channel)

    def get_channel_index(self, channel: str or int):
        """
        Find index for channel using regex. Returns the first matched string.

        Parameters
        ----------
        channel: string or int
            The channel or index to be used
        """
        if isinstance(channel, str):
            channel = find_channel_index(self.channels, channel)
            if channel is None:
                raise Warning(
                    f"Reference channel {channel} not in the available channels: {self.channels}"
                )
        return channel

    @staticmethod
    def ifoob_pad(full, slices):
        """
        Returns the slices padded if it is out of bounds.

        Parameters
        ----------
        full: array
            Slice of OMERO image (zstacks, x, y) - the entire position
            with zstacks as first axis
        slices: tuple of two slices
            Delineates indiceds for the x- and y- ranges of the tile.

        Returns
        -------
        trap: array
            A tile with all z stacks for the given slices.
            If some padding is needed, the median of the image is used.
            If much padding is needed, a tile of NaN is returned.
        """
        # number of pixels in the y direction
        max_size = full.shape[-1]
        # ignore parts of the tile outside of the image
        y, x = [slice(max(0, s.start), min(max_size, s.stop)) for s in slices]
        # get the tile including all z stacks
        trap = full[:, y, x]
        # find extent of padding needed in x and y
        padding = np.array(
            [(-min(0, s.start), -min(0, max_size - s.stop)) for s in slices]
        )
        if padding.any():
            tile_size = slices[0].stop - slices[0].start
            if (padding > tile_size / 4).any():
                # too much of the tile is outside of the image
                # fill with NaN
                trap = np.full((full.shape[0], tile_size, tile_size), np.nan)
            else:
                # pad tile with median value of trap image
                trap = np.pad(trap, [[0, 0]] + padding.tolist(), "median")
        return trap


def find_channel_index(image_channels: t.List[str], channel: str):
    """
    Access
    """
    for i, ch in enumerate(image_channels):
        found = re.match(channel, ch, re.IGNORECASE)
        if found:
            if len(found.string) - (found.endpos - found.start()):
                self._log(f"Channel {channel} matched {ch} using regex")
            return i


def find_channel_name(image_channels: t.List[str], channel: str):
    """
    Find the name of the channel according to a given channel regex.
    """
    index = find_channel_index(image_channels, channel)
    if index is not None:
        return image_channels[index]



#!/usr/bin/env python3
"""
Dataset is a group of classes to manage multiple types of experiments:
 - Remote experiments on an OMERO server (located in src/aliby/io/omero.py)
 - Local experiments in a multidimensional OME-TIFF image containing the metadata
 - Local experiments in a directory containing multiple positions in independent images with or without metadata
"""
import os
import shutil
import time
import typing as t
from abc import ABC, abstractproperty, abstractmethod
from pathlib import Path, PosixPath


from aliby.io.image import ImageLocalOME


def dispatch_dataset(expt_id: int or str, **kwargs):
    """
    Choose a subtype of dataset based on the identifier.

    Input:
    --------
    expt_id: int or string serving as dataset identifier.

    Returns:
    --------
    Callable Dataset instance, either network-dependent or local.
    """
    if isinstance(expt_id, int):  # Is an experiment online

        from aliby.io.omero import Dataset

        return Dataset(expt_id, **kwargs)

    elif isinstance(expt_id, str):  # Files or Dir
        expt_path = Path(expt_id)
        if expt_path.is_dir():
            return DatasetLocalDir(expt_path)
        else:
            return DatasetLocalOME(expt_path)
    else:
        raise Warning("Invalid expt_id")


class DatasetLocalABC(ABC):
    """
    Abstract Base class to fetch local files, either OME-XML or raw images.
    """

    _valid_suffixes = ("tiff", "png")
    _valid_meta_suffixes = ("txt", "log")

    def __init__(self, dpath: t.Union[str, PosixPath], *args, **kwargs):
        self.path = Path(dpath)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False

    @property
    def dataset(self):
        return self.path

    @property
    def name(self):
        return self.path.name

    @property
    def unique_name(self):
        return self.path.name

    @abstractproperty
    def date(self):
        pass

    @property
    def files(self):
        if not hasattr(self, "_files"):
            self._files = {
                f: f
                for f in self.path.rglob("*")
                if any(
                    str(f).endswith(suffix)
                    for suffix in self._valid_meta_suffixes
                )
            }
        return self._files

    def cache_logs(self, root_dir):
        # Copy metadata files to results folder
        for name, annotation in self.files.items():
            shutil.copy(annotation, root_dir / name.name)
        return True

    @abstractmethod
    def get_images(self):
        # Return a dictionary with the name of images and their unique identifiers
        pass


class DatasetLocalDir(DatasetLocalABC):
    """
    Organise an entire dataset, composed of multiple images, as a directory containing directories with individual files.
    It relies on ImageDir to manage images.
    """

    def __init__(self, dpath: t.Union[str, PosixPath], *args, **kwargs):
        super().__init__(dpath)

    @property
    def date(self):
        # Use folder creation date, for cases where metadata is minimal
        return time.strftime(
            "%Y%m%d", time.strptime(time.ctime(os.path.getmtime(self.path)))
        )

    def get_images(self):
        return {
            folder.name: folder
            for folder in self.path.glob("*/")
            if any(
                path
                for suffix in self._valid_suffixes
                for path in folder.glob(f"*.{suffix}")
            )
        }


class DatasetLocalOME(DatasetLocalABC):
    """Load a dataset from a folder

    We use a given image of a dataset to obtain the metadata,
    as we cannot expect folders to contain this information.

    It uses the standard OME-TIFF file format.
    """

    def __init__(self, dpath: t.Union[str, PosixPath], *args, **kwargs):
        super().__init__(dpath)
        assert len(self.get_images()), "No .tiff files found"

    @property
    def date(self):
        # Access the date from the metadata of the first position
        return ImageLocalOME(list(self.get_images().values())[0]).date

    def get_images(self):
        # Fetches all valid formats and overwrites if duplicates with different suffix
        return {
            f.name: str(f)
            for suffix in self._valid_suffixes
            for f in self.path.glob(f"*.{suffix}")
        }



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src',
 'agora': 'src/agora',
 'agora.io': 'src/agora/io',
 'agora.utils': 'src/agora/utils',
 'extraction': 'src/extraction',
 'extraction.core': 'src/extraction/core',
 'extraction.core.functions': 'src/extraction/core/functions',
 'extraction.core.functions.custom': 'src/extraction/core/functions/custom',
 'logfile_parser': 'src/logfile_parser',
 'postprocessor': 'src/postprocessor',
 'postprocessor.benchmarks': 'src/postprocessor/benchmarks',
 'postprocessor.core': 'src/postprocessor/core',
 'postprocessor.core.functions': 'src/postprocessor/core/functions',
 'postprocessor.core.multisignal': 'src/postprocessor/core/multisignal',
 'postprocessor.core.processes': 'src/postprocessor/core/processes',
 'postprocessor.core.reshapers': 'src/postprocessor/core/reshapers',
 'postprocessor.routines': 'src/postprocessor/routines'}

packages = \
['agora',
 'agora.io',
 'agora.utils',
 'aliby',
 'aliby.bin',
 'aliby.io',
 'aliby.lineage',
 'aliby.tile',
 'aliby.track',
 'aliby.utils',
 'extraction',
 'extraction.core',
 'extraction.core.functions',
 'extraction.core.functions.custom',
 'logfile_parser',
 'postprocessor',
 'postprocessor.benchmarks',
 'postprocessor.core',
 'postprocessor.core.functions',
 'postprocessor.core.multisignal',
 'postprocessor.core.processes',
 'postprocessor.core.reshapers',
 'postprocessor.routines']

package_data = \
{'': ['*'], 'logfile_parser': ['grammars/*']}

install_requires = \
['Bottleneck>=1.3.5,<2.0.0',
 'GitPython>=3.1.27,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'aliby-baby>=0.1.15,<0.2.0',
 'dask>=2021.12.0,<2022.0.0',
 'flatten-dict>=0.4.2,<0.5.0',
 'gaussianprocessderivatives>=0.1.5,<0.2.0',
 'h5py==2.10',
 'imageio==2.8.0',
 'leidenalg>=0.8.8,<0.9.0',
 'more-itertools>=8.12.0,<9.0.0',
 'numpy>=1.21.6',
 'opencv-python==4.1.2.30',
 'p-tqdm>=1.3.3,<2.0.0',
 'pandas>=1.3.3',
 'pathos>=0.2.8,<0.3.0',
 'py-find-1st>=1.1.5,<2.0.0',
 'pycatch22>=0.4.2,<0.5.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'scikit-image>=0.18.1',
 'scikit-learn>=1.0.2',
 'scipy>=1.7.3',
 'tqdm>=4.62.3,<5.0.0',
 'xmltodict>=0.13.0,<0.14.0']

extras_require = \
{'network': ['omero-py>=5.6.2', 'zeroc-ice==3.6.5'],
 'omero': ['omero-py>=5.6.2']}

entry_points = \
{'console_scripts': ['aliby-run = aliby.bin.run:run']}

setup_kwargs = {
    'name': 'aliby',
    'version': '0.1.55',
    'description': 'Process and analyse live-cell imaging data',
    'long_description': '# ALIBY (Analyser of Live-cell Imaging for Budding Yeast)\n\n[![docs](https://readthedocs.org/projects/aliby/badge/?version=master)](https://aliby.readthedocs.io/en/latest)\n[![PyPI version](https://badge.fury.io/py/aliby.svg)](https://badge.fury.io/py/aliby)\n[![pipeline](https://gitlab.com/aliby/aliby/badges/master/pipeline.svg?key_text=master)](https://gitlab.com/aliby/aliby/-/pipelines)\n[![dev pipeline](https://gitlab.com/aliby/aliby/badges/dev/pipeline.svg?key_text=dev)](https://gitlab.com/aliby/aliby/-/commits/dev)\n[![coverage](https://gitlab.com/aliby/aliby/badges/dev/coverage.svg)](https://gitlab.com/aliby/aliby/-/commits/dev)\n\nEnd-to-end processing of cell microscopy time-lapses. ALIBY automates segmentation, tracking, lineage predictions, post-processing and report production. It leverages the existing Python ecosystem and open-source scientific software available to produce seamless and standardised pipelines.\n\n## Quickstart Documentation\nInstallation of [VS Studio](https://visualstudio.microsoft.com/downloads/#microsoft-visual-c-redistributable-for-visual-studio-2022) Native MacOS support for is under work, but you can use containers (e.g., Docker, Podman) in the meantime.\n\nFor analysing local data\n ```bash\npip install aliby # aliby[network] if you want to access an OMERO server\n ```\n\nSee our [installation instructions]( https://aliby.readthedocs.io/en/latest/INSTALL.html ) for more details.\n\n### CLI\n\n ```bash\naliby-run --expt_id EXPT_PATH --distributed 4 --tps None\n ```\n\nAnd to run Omero servers, the basic arguments are shown:\n ```bash\n aliby-run --expt_id XXX --host SERVER.ADDRESS --user USER --password PASSWORD \n ```\n\nThe output is a folder with the original logfiles and a set of hdf5 files, one with the results of each multidimensional inside.\n\n## Using specific components\n\n### Access raw data\n\nALIBY\'s tooling can also be used as an interface to OMERO servers, taking care of fetching data when needed.\n ```python\nfrom aliby.io.dataset import Dataset\nfrom aliby.io.image import Image\n\nserver_info= {\n            "host": "host_address",\n            "username": "user",\n            "password": "xxxxxx"}\nexpt_id = XXXX\ntps = [0, 1] # Subset of positions to get.\n\nwith Dataset(expt_id, **server_info) as conn:\n    image_ids = conn.get_images()\n\n#To get the first position\nwith Image(list(image_ids.values())[0], **server_info) as image:\n    dimg = image.data\n    imgs = dimg[tps, image.metadata["channels"].index("Brightfield"), 2, ...].compute()\n    # tps timepoints, Brightfield channel, z=2, all x,y\n```\n\n### Tiling the raw data\n\nA `Tiler` object performs trap registration. It may be built in different ways but the simplest one is using an image and a the default parameters set.\n\n```python\nfrom aliby.tile.tiler import Tiler, TilerParameters\nwith Image(list(image_ids.values())[0], **server_info) as image:\n    tiler = Tiler.from_image(image, TilerParameters.default())\n    tiler.run_tp(0)\n```\n\nThe initialisation should take a few seconds, as it needs to align the images\nin time.\n\nIt fetches the metadata from the Image object, and uses the TilerParameters values (all Processes in aliby depend on an associated Parameters class, which is in essence a dictionary turned into a class.)\n\n#### Get a timelapse for a given trap\n```python\nfpath = "h5/location"\n\ntrap_id = 9\ntrange = list(range(0, 30))\nncols = 8\n\nriv = remoteImageViewer(fpath)\ntrap_tps = riv.get_trap_timepoints(trap_id, trange, ncols)\n```\n\nThis can take several seconds at the moment.\nFor a speed-up: take fewer z-positions if you can.\n\n#### Get the traps for a given time point\nAlternatively, if you want to get all the traps at a given timepoint:\n\n```python\ntimepoint = 0\nseg_expt.get_tiles_timepoints(timepoint, tile_size=96, channels=None,\n                                z=[0,1,2,3,4])\n```\n\n\n### Contributing\nSee [CONTRIBUTING](https://aliby.readthedocs.io/en/latest/INSTALL.html) on how to help out or get involved.\n',
    'author': 'Alan Munoz',
    'author_email': 'alan.munoz@ed.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
